#ifndef __itkekADRevolveImageToImageMetricBase_h
#define __itkekADRevolveImageToImageMetricBase_h


#include "itkImageToImageMetric.h"

#include "vnl/vnl_vector.h"
#include "vnl/vnl_matrix.h"
#include "vnl/vnl_vector_fixed.h"
#include "vnl/vnl_matrix_fixed.h"
#include <utility>
#include <cmath>
#include <string>

#include "RevolveDriver.h"
#include "itkekMacro.h"
#include "ObjectiveFunctionBase.h"
#include "itkVectorLinearInterpolateImageFunction.h"
#include "itkekADRevolveTransformBase.h"

namespace itkek
{



/** \class ADRevolveImageToImageMetricBase
 * \brief Computes similarity between two objects to be registered
 *
 * This class is the same as the itkek::MeanSquaresImageToImageMetric
 * except that it computes derivatives using automatic differentiation.
 * Note: It only works if the transform is an 
 * itkek::ADBSplineDeformableTransform.
 *
 * \ingroup RegistrationMetrics
 */



template < class TFixedImage=itk::Image<float, 3u>, 
	   class TMovingImage=itk::Image<float, 3u> > 
class ITK_EXPORT ADRevolveImageToImageMetricBase : 
   public itk::ImageToImageMetric< TFixedImage, TMovingImage> ,
   public adolc::ObjectiveFunctionBase
{

   public:

      
      /** Standard class typedefs. */
      typedef ADRevolveImageToImageMetricBase    Self;
      typedef itk::ImageToImageMetric< TFixedImage, TMovingImage> Superclass;
      
      typedef itk::SmartPointer<Self>         Pointer;
      typedef itk::SmartPointer<const Self>   ConstPointer;

      /** Method for creation through the object factory. */
      //itkNewMacro(Self);
 
      /** Run-time type information (and related methods). */
      itkTypeMacro(ADRevolveImageToImageMetricBase, 
		   itk::ImageToImageMetric);

 

      /** Types transferred from the base class */
      typedef typename Superclass::RealType                 RealType;
      typedef typename Superclass::TransformType            TransformType;
      typedef typename Superclass::TransformPointer         TransformPointer;
      typedef typename Superclass::TransformParametersType  
                                   TransformParametersType;
      typedef typename Superclass::TransformJacobianType TransformJacobianType;
      typedef typename Superclass::GradientPixelType        GradientPixelType;

      typedef typename Superclass::MeasureType              MeasureType;
      typedef typename Superclass::DerivativeType           DerivativeType;
      typedef typename Superclass::FixedImageType           FixedImageType;
      typedef typename Superclass::MovingImageType          MovingImageType;
      typedef typename Superclass::FixedImageConstPointer   
                                   FixedImageConstPointer;
      typedef typename Superclass::MovingImageConstPointer  
                                   MovingImageConstPointer;

 
      typedef  typename Superclass::InputPointType           InputPointType;
      typedef typename  Superclass::OutputPointType          OutputPointType;
      typedef typename  TransformType::InputVectorType       InputVectorType;
      typedef typename  TransformType::OutputVectorType      OutputVectorType;

      typedef typename Superclass::InterpolatorPointer   InterpolatorPointer;
      typedef typename Superclass::GradientImageType  GradientImageType;
      typedef typename Superclass::GradientImagePointer  GradientImagePointer;

      typedef  itk::VectorLinearInterpolateImageFunction<
         GradientImageType, double >    GradientInterpolatorType;
      typedef typename GradientInterpolatorType::Pointer GradientInterpolatorPointer;

 
      typedef TransformParametersType  ParametersType;

      typedef ADRevolveTransformBase<double,3,3> ADRevolveTransformType;

      typedef adolc::ObjectiveFunctionBase::Label Label;
      typedef adolc::ObjectiveFunctionBase::LabelsContainerType LabelsContainerType;


      void Initialize(void) throw ( itk::ExceptionObject );

      void SetNumberOfSnapshots(int n) const {s_Revolve.SetSnaps(n);}

      virtual void CopyData(const adolc::ObjectiveFunctionBase* rhs);

      // redefine this function from the base class
      virtual void SetTransform (TransformType* _arg);
     

      /** Return the number of parameters required by the Transform */
      unsigned int GetNumberOfParameters(void) const 
      { return m_TransformAD->GetNumberOfParameters(); }


      /**  Get the value for single valued optimizers. */
      MeasureType GetValue( const TransformParametersType & parameters ) const;

      /** Get the derivatives of the match measure. */
      void GetDerivative( const TransformParametersType & parameters,
			  DerivativeType & derivative ) const;

      /**  Get value and derivatives for multiple valued optimizers. */
      void GetValueAndDerivative( const TransformParametersType & parameters,
				  MeasureType& Value, 
				  DerivativeType& Derivative ) const;

      
      
   protected:
      
      ADRevolveImageToImageMetricBase();
      virtual ~ADRevolveImageToImageMetricBase();

      typename ADRevolveTransformType::Pointer  m_TransformAD; // not static!

      static FixedImageConstPointer      s_FixedImage;
      static MovingImageConstPointer     s_MovingImage;
      static InterpolatorPointer         s_Interpolator;
      static GradientImagePointer        s_GradientImage;
      static GradientInterpolatorPointer s_GradientInterpolator;
      static int                         s_NumberOfVoxelsInImage;
      static int                         s_NumberOfVoxelsPerTimeStep;


   private:


      static adolc::Revolve s_Revolve;


      //purposely not implemented:
      ADRevolveImageToImageMetricBase(const Self&); 
      void operator=(const Self&); //purposely not implemented
      
};

}

#ifndef ITK_MANUAL_INSTANTIATION
#include "itkekADRevolveImageToImageMetricBase.txx"
#endif

#endif
