#ifndef __itkekADRevolveBSplineDeformableTransform_h
#define __itkekADRevolveBSplineDeformableTransform_h

#include <iostream>
#include "itkekADRevolveTransformBase.h"
#include "itkImage.h"
#include "itkImageRegion.h"
#include "itkBSplineInterpolationWeightFunction.h"




namespace itkek
{

/** \class ADRevolveBSplineDeformableTransform
 * \brief Deformable transform using a BSpline representation
 *
 * This class is very similar to the class itk::BSplineDeformableTransform
 * except that this one has been modified to allow for automatic
 * differentiation using the ADOLC package. To achieve this several functions
 * were added.
 *
 *
 * \ingroup Transforms
 */
template <
    class TScalarType = double,          // Data type for scalars (float or double)
    unsigned int NDimensions = 3,        // Number of dimensions
    unsigned int VSplineOrder = 3 >      // Spline order
class ITK_EXPORT ADRevolveBSplineDeformableTransform : 
          public ADRevolveTransformBase< TScalarType, NDimensions, NDimensions >
{
public:
  /** Standard class typedefs. */
  typedef ADRevolveBSplineDeformableTransform Self;
  typedef ADRevolveTransformBase< TScalarType, NDimensions, NDimensions > Superclass;
  typedef itk::SmartPointer<Self>        Pointer;
  typedef itk::SmartPointer<const Self>  ConstPointer;
      
  /** New macro for creation of through the object factory.*/
  itkNewMacro( Self );

  /** Run-time type information (and related methods). */
  itkTypeMacro( ADRevolveBSplineDeformableTransform, ADRevolveTransformBase );

  /** Dimension of the domain space. */
  itkStaticConstMacro(SpaceDimension, unsigned int, NDimensions);

  /** The BSpline order. */
  itkStaticConstMacro(SplineOrder, unsigned int, VSplineOrder);

  /** Standard scalar type for this class. */
  typedef typename Superclass::ScalarType ScalarType;

  /** Standard parameters container. */
  typedef typename Superclass::ParametersType ParametersType;

  /** Standard Jacobian container. */
  typedef typename Superclass::JacobianType JacobianType;

  /** Standard vector type for this class. */
  typedef itk::Vector<TScalarType,
                 itkGetStaticConstMacro(SpaceDimension)> InputVectorType;
  typedef itk::Vector<TScalarType,
                 itkGetStaticConstMacro(SpaceDimension)> OutputVectorType;

  /** Standard covariant vector type for this class. */
  typedef itk::CovariantVector<TScalarType,
                          itkGetStaticConstMacro(SpaceDimension)> InputCovariantVectorType;
  typedef itk::CovariantVector<TScalarType,
                          itkGetStaticConstMacro(SpaceDimension)> OutputCovariantVectorType;
  
  /** Standard vnl_vector type for this class. */
  typedef vnl_vector_fixed<TScalarType,
                           itkGetStaticConstMacro(SpaceDimension)> InputVnlVectorType;
  typedef vnl_vector_fixed<TScalarType,
                           itkGetStaticConstMacro(SpaceDimension)> OutputVnlVectorType;
  
  /** Standard coordinate point type for this class. */
  typedef itk::Point<TScalarType,
                itkGetStaticConstMacro(SpaceDimension)> InputPointType;
  typedef itk::Point<TScalarType,
                itkGetStaticConstMacro(SpaceDimension)> OutputPointType;
  
  /** This method sets the parameters of the transform.
   * For a BSpline deformation transform, the parameters are the BSpline 
   * coefficients on a sparse grid. 
   * 
   * The parameters are N number of N-D grid of coefficients. Each N-D grid 
   * is represented as a flat array of doubles 
   * (in the same configuration as an itk::Image).
   * The N arrays are then concatenated to form one parameter array.
   *
   * For efficiency, this transform does not make a copy of the parameters.
   * It only keeps a pointer to the input parameters. It assumes that the memory
   * is managed by the caller.
   *
   * This method wraps each grid as itk::Image's using the user specified
   * grid region, spacing and origin.
   * NOTE: The grid region, spacing and origin must be set first.
   *
   */
  void SetParameters(const ParametersType & parameters);

  /** This method can ONLY be invoked AFTER calling SetParameters(). 
   *  This restriction is due to the fact that the BSplineDeformableTransform
   *  does not copy the array of paramters internally, instead it keeps a 
   *  pointer to the user-provided array of parameters. This method is also
   *  in violation of the const-correctness of the parameters since the 
   *  parameter array has been passed to the transform on a 'const' basis but
   *  the values get modified when the user invokes SetIdentity().
   */
  void SetIdentity();

  /** Get the Transformation Parameters. */
  virtual const ParametersType& GetParameters(void) const;

  /** Parameters as SpaceDimension number of images. */
  typedef typename ParametersType::ValueType PixelType;
  typedef itk::Image<PixelType,itkGetStaticConstMacro(SpaceDimension)> ImageType;
  typedef typename ImageType::Pointer ImagePointer;

  /** Get the array of coefficient images. */
  virtual ImagePointer * GetCoefficientImage()
  { return m_CoefficientImage; }

  /** Set the array of coefficient images.
   *
   * This is an alternative API for setting the BSpline coefficients
   * as an array of SpaceDimension images. The grid region spacing 
   * and origin is taken from the first image. It is assume that
   * the buffered region of all the subsequent images are the same 
   * as the first image. Note that no error checking is done.
   *
   * Warning: use either the SetParameters() or SetCoefficientImage()
   * API. Mixing the two modes may results in unexpected results.
   *
   */
  virtual void SetCoefficientImage( ImagePointer images[] );  

  /** Typedefs for specifying the extend to the grid. */
  typedef itk::ImageRegion<itkGetStaticConstMacro(SpaceDimension)>    RegionType;
  typedef typename RegionType::IndexType IndexType;
  typedef typename RegionType::SizeType  SizeType;
  typedef typename ImageType::SpacingType SpacingType;
  typedef typename ImageType::PointType OriginType;

  /** This method specifies the region over which the grid resides. */
  virtual void SetGridRegion( const RegionType& region );
  itkGetMacro( GridRegion, RegionType );

  /** This method specifies the grid spacing or resolution. */
  virtual void SetGridSpacing( const SpacingType& spacing );
  itkGetMacro( GridSpacing, SpacingType );

  /** This method specifies the grid origin. */
  virtual void SetGridOrigin( const OriginType& origin );
  itkGetMacro( GridOrigin, OriginType );

  /** Typedef of the bulk transform. */
  typedef itk::Transform<ScalarType,itkGetStaticConstMacro(SpaceDimension),
                    itkGetStaticConstMacro(SpaceDimension)> BulkTransformType;
  typedef typename BulkTransformType::ConstPointer  BulkTransformPointer;

  /** This method specifies the bulk transform to be applied. 
   * The default is the identity transform.
   */
  //itkSetConstObjectMacro( BulkTransform, BulkTransformType );
  virtual void SetBulkTransform (const BulkTransformType* _arg) 
  { 
     itkDebugMacro("setting " << "BulkTransform to " << _arg ); 
     if (this->m_BulkTransform != _arg) 
     { 
	   this->m_BulkTransform = _arg; 
	   this->Modified(); 
     } 
     s_BulkTransform = m_BulkTransform;
  }

  itkGetConstObjectMacro( BulkTransform, BulkTransformType );

  /** Transform points by a BSpline deformable transformation. */
  OutputPointType  TransformPoint(const InputPointType  &point ) const;

  /** Interpolation weights function type. */
  typedef itk::BSplineInterpolationWeightFunction<ScalarType,
                                             itkGetStaticConstMacro(SpaceDimension),
                                             itkGetStaticConstMacro(SplineOrder)> WeightsFunctionType;
  typedef typename WeightsFunctionType::WeightsType WeightsType;
  typedef typename WeightsFunctionType::ContinuousIndexType ContinuousIndexType;

  /** Parameter index array type. */
  typedef itk::Array<unsigned long> ParameterIndexArrayType;

  /** EGK: I've commented out this function */
  /* Transform points by a BSpline deformable transformation. 
   * On return, weights contains the interpolation weights used to compute the 
   * deformation and indices of the x (zeroth) dimension coefficient parameters
   * in the support region used to compute the deformation.
   * Parameter indices for the i-th dimension can be obtained by adding
   * ( i * this->GetNumberOfParametersPerDimension() ) to the indices array.
   */
/*  virtual void TransformPoint( const InputPointType & inputPoint,
                       OutputPointType & outputPoint,
                       WeightsType & weights,
                       ParameterIndexArrayType & indices, 
                       bool & inside ) const;
*/

  /** Get number of weights. */
  unsigned long GetNumberOfWeights() const
  { return m_WeightsFunction->GetNumberOfWeights(); }

  /**  Method to transform a vector - not applicable for this type of transform. */
  virtual OutputVectorType TransformVector(const InputVectorType &) const
  { 
      itkExceptionMacro(<< "Method not applicable for deformable transform." );
      return OutputVectorType(); 
  }

  /**  Method to transform a vnl_vector - not applicable for this type of transform */
  virtual OutputVnlVectorType TransformVector(const InputVnlVectorType &) const
  { 
      itkExceptionMacro(<< "Method not applicable for deformable transform. ");
      return OutputVnlVectorType(); 
  }

  /**  Method to transform a CovariantVector - not applicable for this type of transform */
  virtual OutputCovariantVectorType TransformCovariantVector(
	  const InputCovariantVectorType &) const
  { 
      itkExceptionMacro(<< "Method not applicable for deformable transfrom. ");
      return OutputCovariantVectorType(); 
  } 
    
  /** Compute the Jacobian Matrix of the transformation at one point */
  virtual const JacobianType& GetJacobian(const InputPointType  &point ) const;

  /** Return the number of parameters that completely define the Transfom */
  virtual unsigned int GetNumberOfParameters(void) const;

  /** Return the number of parameters per dimension */
  unsigned int GetNumberOfParametersPerDimension(void) const;

  /** Return the region of the grid wholly within the support region */
  itkGetConstReferenceMacro( ValidRegion, RegionType );



  /** EGK: These are the new functions which I've added to allow for automatic
      differention. */
  virtual void TransformPointAD( const InputPointType& thisPoint, 
				 adoublev& transformedPoint ) const;

  void SetParametersAD(const adouble* parameters);

  /** EGK: In addition, I've added two functions for loading and saving the
      transform from disk */
  void SaveTransform(const char* filename) const;
  void LoadTransform(const char* filename);


  /** EGK: I've also added a function to get the parameters from the base
      class. This is usefull in situations where you need a (global) container
      for holding the bspline parameters but none is readily available. Now the
      base class has an object called m_Parameters which is not used by the
      bspline class, but there is no function available for accessing it (See
      the file itkTransform.h. Also the GetParameters() function in this file
      does not return it either). Therefore we provide one here. */
  ParametersType& GetParametersFromBase(void) const
  { return this->m_Parameters; };


protected:
  /** Print contents of an BSplineDeformableTransform. */
  void PrintSelf(std::ostream &os, itk::Indent indent) const;


  ADRevolveBSplineDeformableTransform();
  virtual ~ADRevolveBSplineDeformableTransform();

  /** Allow subclasses to access and manipulate the weights function. */
  itkSetObjectMacro( WeightsFunction, WeightsFunctionType );
  itkGetObjectMacro( WeightsFunction, WeightsFunctionType );

private:
  ADRevolveBSplineDeformableTransform(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  /** The bulk transform. */
  BulkTransformPointer  m_BulkTransform;

  /** Variables defining the coefficient grid extend. */
  RegionType    m_GridRegion;
  SpacingType   m_GridSpacing;
  OriginType    m_GridOrigin;
  
  RegionType    m_ValidRegion;

  /** Variables defining the interpolation support region. */
  unsigned long m_Offset;
  bool          m_SplineOrderOdd;
  SizeType      m_SupportSize;
  IndexType     m_ValidRegionLast;
  
  /** Array holding images wrapped from the flat parameters. */
  ImagePointer   m_WrappedImage[NDimensions];

  /** Array of images representing the B-spline coefficients in each dimension. */
  ImagePointer   m_CoefficientImage[NDimensions];

  /** Jacobian as SpaceDimension number of images. */
  typedef typename JacobianType::ValueType JacobianPixelType;
  typedef itk::Image<JacobianPixelType,
                itkGetStaticConstMacro(SpaceDimension)> JacobianImageType;
 
  typename JacobianImageType::Pointer m_JacobianImage[NDimensions];

  /** Keep track of last support region used in computing the Jacobian
   * for fast resetting of Jacobian to zero.
   */
  mutable IndexType m_LastJacobianIndex;

  /** Keep a pointer to the input parameters. */
  const ParametersType *  m_InputParametersPointer;

  /** Pointer to function used to compute Bspline interpolation weights. */
  typename WeightsFunctionType::Pointer  m_WeightsFunction;

  /** Check if a continuous index is inside the valid region. */
  bool InsideValidRegion( const ContinuousIndexType& index ) const;




  /** adouble pointer to input parameters */
  static const adouble *       s_InputParametersPointerAD;

  static BulkTransformPointer  s_BulkTransform;
  static ImagePointer          s_CoefficientImage[NDimensions];
  static SpacingType           s_GridSpacing;
  static OriginType            s_GridOrigin;
  static RegionType            s_GridRegion;
  static ParametersType        s_DummyParameters;
  static RegionType            s_ValidRegion;
  static IndexType             s_ValidRegionLast;

  bool InsideValidRegionAD( const ContinuousIndexType& index ) const;

}; //class ADRevolveBSplineDeformableTransform


}  // namespace itkek


#ifndef ITK_MANUAL_INSTANTIATION
#include "itkekADRevolveBSplineDeformableTransform.txx"
#endif

#endif /* __itkekADRevolveBSplineDeformableTransform_h */
