#ifndef __itkekADRevolveEuler3DTransform_h
#define __itkekADRevolveEuler3DTransform_h


#include "itkekADRevolveMatrixOffsetTransformBase.h"

namespace itkek
{

/** \brief ADRevolveEuler3DTransform of a vector space (e.g. space coordinates)
 *
 * This transform applies a rotation and translation to the space given 3 euler
 * angles and a 3D offset.
 *
 * \ingroup Transforms
 */
template < class TScalarType=double >    // Data type for scalars (float or double)
class ITK_EXPORT ADRevolveEuler3DTransform : 
            public ADRevolveMatrixOffsetTransformBase< TScalarType,3,3 > 
{
public:
  /** Standard class typedefs. */
  typedef ADRevolveEuler3DTransform                  Self;
  typedef ADRevolveMatrixOffsetTransformBase< TScalarType,3,3 >   Superclass;
  typedef itk::SmartPointer<Self>                Pointer;
  typedef itk::SmartPointer<const Self>          ConstPointer;
    
  /** New macro for creation of through a Smart Pointer. */
  itkNewMacro( Self );

  /** Run-time type information (and related methods). */
  itkTypeMacro( ADRevolveEuler3DTransform, ADRevolveMatrixOffsetTransformBase );

  /** Dimension of the space. */
  itkStaticConstMacro(SpaceDimension, unsigned int, 3);
  itkStaticConstMacro(InputSpaceDimension, unsigned int, 3);
  itkStaticConstMacro(OutputSpaceDimension, unsigned int, 3);
  itkStaticConstMacro(ParametersDimension, unsigned int, 6);

  typedef typename Superclass::ParametersType             ParametersType;
  typedef typename Superclass::JacobianType               JacobianType;
  typedef typename Superclass::ScalarType                 ScalarType;
  typedef typename Superclass::InputVectorType            InputVectorType;
  typedef typename Superclass::OutputVectorType           OutputVectorType;
  typedef typename Superclass::InputCovariantVectorType  
                                                     InputCovariantVectorType;
  typedef typename Superclass::OutputCovariantVectorType  
                                                     OutputCovariantVectorType;
  typedef typename Superclass::InputVnlVectorType         InputVnlVectorType;
  typedef typename Superclass::OutputVnlVectorType        OutputVnlVectorType;
  typedef typename Superclass::InputPointType             InputPointType;
  typedef typename Superclass::OutputPointType            OutputPointType;
  typedef typename Superclass::MatrixType                 MatrixType;
  typedef typename Superclass::InverseMatrixType          InverseMatrixType;
  typedef typename Superclass::CenterType                 CenterType;
  typedef typename Superclass::TranslationType            TranslationType;
  typedef typename Superclass::OffsetType                 OffsetType;

  typedef typename Superclass::ScalarType                 AngleType;
  
  /** Set/Get the transformation from a container of parameters
   * This is typically used by optimizers.  There are 6 parameters. The first
   * three represent the angles to rotate around the coordinate axis, and the
   * last three represents the offset. */
  void SetParameters( const ParametersType & parameters );
  const ParametersType& GetParameters(void) const;

  /** Set the rotational part of the transform. */
  void SetRotation(ScalarType angleX,ScalarType angleY,ScalarType angleZ);
  itkGetConstMacro(AngleX, ScalarType);
  itkGetConstMacro(AngleY, ScalarType);
  itkGetConstMacro(AngleZ, ScalarType);

  /** This method computes the Jacobian matrix of the transformation.
   * given point or vector, returning the transformed point or
   * vector. The rank of the Jacobian will also indicate if the 
   * transform is invertible at this point. */
  const JacobianType & GetJacobian(const InputPointType  &point ) const;

  /** Set/Get the order of the computation. Default ZXY */
  itkSetMacro(ComputeZYX,bool);
  itkGetConstMacro(ComputeZYX,bool);

  virtual void SetIdentity(void);

  virtual unsigned int GetNumberOfParameters(void) const { return 6; }


  // These are two new functions which I've added to allow for automatic
  // differention. 
  // No need to redefine. Same as in base class
  //adoublev&  TransformPointAD( const InputPointType &point );

  void SetParametersAD(const adouble* parameters);

protected:
  ADRevolveEuler3DTransform();
//  ADRevolveEuler3DTransform(const MatrixType & matrix,
//                   const OutputPointType & offset);
//  ADRevolveEuler3DTransform(unsigned int outputSpaceDims,
//                   unsigned int paramsSpaceDims);

  ~ADRevolveEuler3DTransform(){};

  void PrintSelf(std::ostream &os, itk::Indent indent) const;

  void SetVarRotation(ScalarType angleX, ScalarType angleY, ScalarType angleZ)
  { m_AngleX = angleX; m_AngleY = angleY; m_AngleZ = angleZ; };

  /** Compute the components of the rotation matrix in the superclass. */
  void ComputeMatrix(void);
  void ComputeMatrixParameters(void);

private:
  ADRevolveEuler3DTransform(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  ScalarType  m_AngleX; 
  ScalarType  m_AngleY; 
  ScalarType  m_AngleZ;
  bool        m_ComputeZYX;



      // The following are all temporary variables and therefore don't need to
      // be declared static.
      //static adoublem tmp,RotationX,RotationY,RotationZ;
      //static adouble angleXAD,angleYAD,angleZAD;
      //static adouble cx,sx,cy,sy,cz,sz;

      void MatrixMultiply(adoublem& r, const adoublem& m1, 
			  const adoublem& m2) const
      {
	  for(int i=0;i<3;i++)
	    for(int j=0;j<3;j++)
	    {
	       r[i][j] = 0.0;
	       for(int n=0;n<3;n++)
	       {
		   r[i][j] += m1[i][n]*m2[n][j];
	       }
	    }
      }
      
      void TripleMatrixMultiply(adoublem& r, const adoublem& m1, 
				const adoublem& m2, const adoublem& m3,
				adoublem& tmp) const
      {
	  //static adoublem tmp(3,3);
	  MatrixMultiply(tmp,m1,m2);
	  MatrixMultiply(r,tmp,m3);
      }


}; //class ADRevolveEuler3DTransform


}  // namespace itkek


#ifndef ITK_MANUAL_INSTANTIATION
#include "itkekADRevolveEuler3DTransform.txx"
#endif

#endif /* __itkekADRevolveEuler3DTransform_h */
