#ifndef __itkekADRevolveTransformBase_h
#define __itkekADRevolveTransformBase_h


#include "ObjectiveFunctionBase.h"
#include "itkTransform.h"


namespace itkek
{


/**
   This class represents the objective function used for non-rigid
   registration. 
 */
template <class TScalarType, // Data type for scalars (float or double)
          unsigned int NInputDimensions=3, 
          unsigned int NOutputDimensions=3>
class ITK_EXPORT ADRevolveTransformBase : 
   public itk::Transform<double,3, 3>
{

   public:

      typedef ADRevolveTransformBase    Self;
      typedef itk::Transform<TScalarType,NInputDimensions,NOutputDimensions> Superclass;
      
      typedef itk::SmartPointer<Self>         Pointer;
      typedef itk::SmartPointer<const Self>   ConstPointer;
      

      /** Run-time type information (and related methods). */
      itkTypeMacro( ADRevolveTransformBase, itk::Transform );

      typedef itk::Point<double,3> PointType;

      typedef PointType InputPointType;
      typedef PointType OutputPointType;



      typedef adoublev ADVectorType;

      typedef adolc::ObjectiveFunctionBase::Label Label;
      typedef adolc::ObjectiveFunctionBase::LabelsContainerType LabelsContainerType;


      virtual void SetParametersAD(const adouble* ptr) = 0;

      virtual void TransformPointAD( const InputPointType& thisPoint, 
				     adoublev& transformedPoint ) const = 0;


      // If the subclass has more than one time step, then the next function
      // MAY need to be overridden
      virtual void CopyData(const Self* rhs) {}


      // If the subclass has more than one time step, then the next two
      // functions MUST be overridden.
      virtual void Function(Label& label)
      { itkExceptionMacro( << "Subclasses should override this method" ); };

      virtual LabelsContainerType GetLabels() const
      { itkExceptionMacro( << "Subclasses should override this method" );
      return LabelsContainerType(); };

      virtual void EnableCheckpointing() { s_CheckpointingFlag = true; }

      virtual void DisableCheckpointing() { s_CheckpointingFlag = false; }


      // If the subclass has more than one time step, override it in the
      // subclass so that it returns true.
      virtual bool DoesTransformHaveMoreThanOneTimeStep() const {return false;}


      

   protected:


      //ADRevolveTransformBase();
      ADRevolveTransformBase(unsigned int Dimension, unsigned int NumberOfParameters);
      virtual ~ADRevolveTransformBase();


      static bool s_CheckpointingFlag;


   private:
      ADRevolveTransformBase(const ADRevolveTransformBase& rhs); // not imple
      ADRevolveTransformBase& operator=(const ADRevolveTransformBase& rhs); //not imple



};

// usefull macro
#define TIMESTEP(N) { if(s_CheckpointingFlag) {return; L ## N:;} }

/* Example: If N is 3, then this expands to:
{
  if(m_CheckpointingFlag)
  {
    return;
    L3:;
  }
}
*/


}

#ifndef ITK_MANUAL_INSTANTIATION
#include "itkekADRevolveTransformBase.txx"
#endif

#endif
